@@ -153,14 +153,19 @@ static gboolean xfconf_backend_perchannel_xml_exists(XfconfBackend *backend,
                                                      const gchar *property,
                                                      gboolean *exists,
                                                      GError **error);
-static gboolean xfconf_backend_perchannel_xml_remove(XfconfBackend *backend,
-                                                     const gchar *channel,
-                                                     const gchar *property,
-                                                     gboolean recursive,
-                                                     GError **error);
+static gboolean xfconf_backend_perchannel_xml_reset(XfconfBackend *backend,
+                                                    const gchar *channel,
+                                                    const gchar *property,
+                                                    gboolean recursive,
+                                                    GError **error);
 static gboolean xfconf_backend_perchannel_xml_list_channels(XfconfBackend *backend,
                                                             GSList **channels,
                                                             GError **error);
+static gboolean xfconf_backend_perchannel_xml_is_property_locked(XfconfBackend *backend,
+                                                                 const gchar *channel,
+                                                                 const gchar *property,
+                                                                 gboolean *locked,
+                                                                 GError **error);
 static gboolean xfconf_backend_perchannel_xml_flush(XfconfBackend *backend,
                                                     GError **error);
 static void xfconf_backend_perchannel_xml_register_property_changed_func(XfconfBackend *backend,
@@ -190,6 +195,9 @@ static GNode *xfconf_proptree_lookup_node(GNode *proptree,
 static gboolean xfconf_proptree_remove(GNode *proptree,
                                        const gchar *name);
 static void xfconf_proptree_destroy(GNode *proptree);
+static gchar *xfconf_proptree_build_propname(GNode *prop_node,
+                                             gchar *buf,
+                                             gsize buflen);
 
 static void xfconf_property_free(XfconfProperty *property);
 
@@ -244,8 +252,9 @@ xfconf_backend_perchannel_xml_backend_init(XfconfBackendInterface *iface)
     iface->get = xfconf_backend_perchannel_xml_get;
     iface->get_all = xfconf_backend_perchannel_xml_get_all;
     iface->exists = xfconf_backend_perchannel_xml_exists;
-    iface->remove = xfconf_backend_perchannel_xml_remove;
+    iface->reset = xfconf_backend_perchannel_xml_reset;
     iface->list_channels = xfconf_backend_perchannel_xml_list_channels;
+    iface->is_property_locked = xfconf_backend_perchannel_xml_is_property_locked;
     iface->flush = xfconf_backend_perchannel_xml_flush;
     iface->register_property_changed_func = xfconf_backend_perchannel_xml_register_property_changed_func;
 }
@@ -316,13 +325,14 @@ xfconf_backend_perchannel_xml_set(XfconfBackend *backend,
             return FALSE;
         }
 
+        if(_xfconf_gvalue_is_equal(&cur_prop->value, value))
+            return TRUE;
+
         if(G_IS_VALUE(&cur_prop->value))
             g_value_unset(&cur_prop->value);
         g_value_copy(value, g_value_init(&cur_prop->value,
                                          G_VALUE_TYPE(value)));
 
-        /* FIXME: this will trigger if the value is replaced by the same
-         * value */
         if(xbpx->prop_changed_func)
             xbpx->prop_changed_func(backend, channel, property, xbpx->prop_changed_data);
     } else {
@@ -492,22 +502,29 @@ xfconf_backend_perchannel_xml_exists(XfconfBackend *backend,
 
 typedef struct
 {
-    XfconfBackend *backend;
+    XfconfBackendPerchannelXml *xbpx;
     const gchar *channel;
 } PropChangeData;
 
-static void nodes_do_propchange_remove(GNode *node,
-                                       gpointer data)
+static gboolean
+nodes_do_propchange_remove(GNode *node,
+                           gpointer data)
 {
-    XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(data);
     PropChangeData *pdata = data;
     XfconfProperty *prop = node->data;
+    gchar prop_fullname[4096];
 
-    if(!G_VALUE_TYPE(&prop->value))
-        return;
 
-    xbpx->prop_changed_func(pdata->backend, pdata->channel,
-                            prop->name, xbpx->prop_changed_data);
+    if(G_VALUE_TYPE(&prop->value)) {
+        pdata->xbpx->prop_changed_func(XFCONF_BACKEND(pdata->xbpx),
+                                       pdata->channel,
+                                       xfconf_proptree_build_propname(node,
+                                                                      prop_fullname,
+                                                                      sizeof(prop_fullname)),
+                                       pdata->xbpx->prop_changed_data);
+    }
+
+    return FALSE;
 }
 
 static gboolean
@@ -533,10 +550,10 @@ do_remove_channel(XfconfBackend *backend,
     if(xbpx->prop_changed_func) {
         PropChangeData pdata;
 
-        pdata.backend = backend;
+        pdata.xbpx = xbpx;
         pdata.channel = channel;
-        g_node_children_foreach(properties, G_TRAVERSE_ALL,
-                                nodes_do_propchange_remove, &pdata);
+        g_node_traverse(properties, G_POST_ORDER, G_TRAVERSE_ALL, -1,
+                        nodes_do_propchange_remove, &pdata);
     }
     g_tree_remove(xbpx->channels, channel);
 
@@ -557,11 +574,11 @@ do_remove_channel(XfconfBackend *backend,
 }
 
 static gboolean
-xfconf_backend_perchannel_xml_remove(XfconfBackend *backend,
-                                     const gchar *channel,
-                                     const gchar *property,
-                                     gboolean recursive,
-                                     GError **error)
+xfconf_backend_perchannel_xml_reset(XfconfBackend *backend,
+                                    const gchar *channel,
+                                    const gchar *property,
+                                    gboolean recursive,
+                                    GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
     GNode *properties = g_tree_lookup(xbpx->channels, channel);
@@ -588,7 +605,6 @@ xfconf_backend_perchannel_xml_remove(XfconfBackend *backend,
             xbpx->prop_changed_func(backend, channel, property, xbpx->prop_changed_data);
     } else {
         GNode *top;
-        XfconfProperty *prop;
         
         if(property[0] && property[1]) {
             PropChangeData pdata;
@@ -605,16 +621,12 @@ xfconf_backend_perchannel_xml_remove(XfconfBackend *backend,
                 return FALSE;
             }
 
-            g_node_unlink(top);
-            prop = top->data;
-            if(G_VALUE_TYPE(&prop->value) && xbpx->prop_changed_func)
-                xbpx->prop_changed_func(backend, channel, property, xbpx->prop_changed_data);
-
-            pdata.backend = backend;
+            pdata.xbpx = xbpx;
             pdata.channel = channel;
-            g_node_children_foreach(top, G_TRAVERSE_ALL,
-                                    nodes_do_propchange_remove, &pdata);
+            g_node_traverse(top, G_POST_ORDER, G_TRAVERSE_ALL, -1,
+                            nodes_do_propchange_remove, &pdata);
 
+            g_node_unlink(top);
             xfconf_proptree_destroy(top);
         } else {
             /* remove the entire channel */
@@ -657,6 +669,30 @@ xfconf_backend_perchannel_xml_list_channels(XfconfBackend *backend,
     return TRUE;
 }
 
+static gboolean
+xfconf_backend_perchannel_xml_is_property_locked(XfconfBackend *backend,
+                                                 const gchar *channel,
+                                                 const gchar *property,
+                                                 gboolean *locked,
+                                                 GError **error)
+{
+    XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
+    GNode *properties;
+    XfconfProperty *prop;
+
+    properties = g_tree_lookup(xbpx->channels, channel);
+    if(!properties) {
+        properties = xfconf_backend_perchannel_xml_load_channel(xbpx, channel,
+                                                                error);
+        if(!properties)
+            return FALSE;
+    }
+
+    prop = xfconf_proptree_lookup(properties, property);
+    *locked = prop ? prop->locked : FALSE;
+    return TRUE;
+}
+
 static gboolean
 xfconf_backend_perchannel_xml_flush(XfconfBackend *backend,
                                     GError **error)
@@ -834,6 +870,33 @@ xfconf_proptree_destroy(GNode *proptree)
     }
 }
 
+static gchar *
+xfconf_proptree_build_propname(GNode *prop_node,
+                               gchar *buf,
+                               gsize buflen)
+{
+    GSList *components = NULL, *lp;
+    GNode *cur;
+
+    for(cur = prop_node; cur; cur = cur->parent) {
+        XfconfProperty *prop = cur->data;
+        if(!prop->name[1])  /* we've hit "/" */
+            break;
+        components = g_slist_prepend(components, prop->name);
+    }
+
+    /* FIXME: optimise */
+    buf[0] = 0;
+    for(lp = components; lp; lp = lp->next) {
+        g_strlcat(buf, "/", buflen);
+        g_strlcat(buf, (gchar *)lp->data, buflen);
+    }
+
+    g_slist_free(components);
+
+    return buf;
+}
+
 
 
 static void
@@ -1562,7 +1625,7 @@ xfconf_format_xml_tag(GString *elem_str,
                     GValue *value1 = g_ptr_array_index(arr, i);
                     gboolean dummy;
 
-                    g_string_append_printf(elem_str, "%s  <value ", spaces);
+                    g_string_append_printf(elem_str, "%s  <value", spaces);
                     if(!xfconf_format_xml_tag(elem_str, value1, TRUE, spaces,
                                               &dummy))
                     {
